"""All function related decorators and objects including inputs and outputs."""
from .inputs import Inputs  # expose for easy import
from .outputs import Outputs  # expose for easy import
from .base import Function, command
